/*
    SPDX-FileCopyrightText: 2020 Albert Astals Cid <aacid@kde.org>

    SPDX-License-Identifier: GPL-2.0-or-later
*/

#include <QString>

namespace Okular
{
void removeRCFileIfVersionSmallerThan(const QString &filePath, int version);
}
